# -*- coding: utf-8 -*-

import time
from random import randint
from functools import wraps

from ...utils import pyv
from .exceptions import ReRaiseException


def re_raise_exc(callback=None, exc_cls=ReRaiseException, message=None):
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as e:
                orig_message = pyv.get_exc_message(e)

                if message:
                    new_message = u'{}{}'.format(
                        message, u' (from {}: {})'.format(e.__class__.__name__, orig_message)
                        if orig_message else
                        '(from {})'.format(e.__class__.__name__)
                    )
                else:
                    new_message = orig_message

                raise exc_cls(new_message)

        return wrapped

    if callable(callback):
        return wrapper(callback)

    return wrapper


def random_file_name(file_ex=None):
    file_ex = file_ex or ''
    file_name = str(
        int(time.time() + randint(0, 1000)),
    )
    file_name += file_ex
    return file_name


def change_name_from_python_to_html(name):
    name = name.replace('_', '-')
    if name.startswith('-'):
        return name[1:]
    if name.endswith('-'):
        return name[:-1]
    return name


def is_ready_state_complete(browser):
    state = browser.execute_script(
        'return document.readyState',
    )
    return state == 'complete'



# -*- coding: utf-8 -*-

from ... import suite
from .case import SeleniumCase
from .extension import EX_NAME


class SeleniumSuiteLayer(suite.SuiteLayer):

    def on_require(self, require):
        if EX_NAME not in require:
            require.append(EX_NAME)


class SeleniumSuite(suite.Suite):

    __case_class__ = SeleniumCase
    __layers__ = (SeleniumSuiteLayer(), )



# -*- coding: utf-8 -*-

"""
Alternative for test runner.
Combine runnable objects and run their.
"""

from __future__ import absolute_import


def get_pool_size_of_value(value, in_two=False):
    from multiprocessing import cpu_count

    if value <= 0:
        size = cpu_count()
    else:
        size = value

    if in_two and value <= 0:
        return int(round(size / 2)) or 2

    return size



# -*- coding: utf-8 -*-

"""
A powerful, extensible and easy to use framework for test development

By Trifonov Mikhail

For support, use the https://github.com/trifonovmixail/seismograph/issues tracker
"""

from .case import skip
from .case import Case
from .case import flows
from .case import skip_if
from .case import assertion
from .case import CaseLayer
from .case import skip_unless
from .case import AssertionBase

from .suite import Suite
from .suite import SuiteLayer

from .script import Script
from .script import AfterScript
from .script import BeforeScript

from .program import main
from .program import Program
from .program import ProgramLayer

from .config import get_config_path_by_env

from .steps import step

from .datastructures import Context


__version__ = '0.2.3'


__all__ = (
    'main',
    'skip',
    'step',
    'Case',
    'flows',
    'Suite',
    'skip_if',
    'Context',
    'Program',
    'assertion',
    'CaseLayer',
    'SuiteLayer',
    'AfterScript',
    'skip_unless',
    'ProgramLayer',
    'BeforeScript',
    'AssertionBase',
    'get_config_path_by_env',
)


from .utils.pyv import check_py_version

check_py_version()

del check_py_version



# -*- coding: utf-8 -*-

from . import runnable


class Script(runnable.RunnableObject):

    __run_point__ = None
    __create_reason__ = False

    def __init__(self, program):
        super(Script, self).__init__()

        self.__program = program

    @property
    def config(self):
        return self.__program.config


class BeforeScript(Script):

    __run_point__ = 'before'


class AfterScript(Script):

    __run_point__ = 'after'



# -*- coding: utf-8 -*-

import logging


logger = logging.getLogger(__name__)


TO_INIT = []


try:
    from . import mocker
    TO_INIT.append(mocker)
except ImportError:
    pass

try:
    from . import alchemy
    TO_INIT.append(alchemy)
except ImportError:
    pass

try:
    from . import selenium
    TO_INIT.append(selenium)
except ImportError:
    pass


logger.debug('Available extensions: {}'.format(TO_INIT))



# -*- coding: utf-8 -*-

import re
from contextlib import contextmanager

from flask import Response
from flask import abort as _abort

from ...utils import pyv

if pyv.IS_PYTHON_2:
    import urlparse
else:
    from urllib import parse


DEFAULT_STATUS_CODE = 200
DEFAULT_HTTP_METHOD = 'GET'

ABORT_ATTRIBUTE_NAME = '__abort__'


def on_file(mock, fp):
    mock.__on_file__(fp)


def abort(mock, status_code):
    setattr(mock, ABORT_ATTRIBUTE_NAME, status_code)


def parse_for_meta(string):
    http_method, status_code, url_rule = [
        item.strip() for item in string.split(' ')
    ]
    return http_method, status_code, url_rule


def parse_for_header(string):
    key = string[:string.index(':'):]
    value = string[string.index(':') + 1:]

    return key.strip(), value.strip()


class BaseMock(object):

    __headers__ = None
    __mime_type__ = None
    __content_type__ = None

    HEADER_REGEXP = re.compile(r'[A-z -]+:.*\s')
    META_REGEXP = re.compile(r'[A-Z]{3,6}\s[0-9]{3}\s\/.*\s')

    def __init__(self,
                 endpoint,
                 url_rule,
                 body=None,
                 headers=None,
                 mime_type=None,
                 status_code=None,
                 http_method=None,
                 content_type=None):
        self._body = body or ''
        self._url_rule = url_rule
        self._endpoint = endpoint
        self._mime_type = mime_type
        self._headers = headers or {}
        self._content_type = content_type
        self._status_code = status_code or DEFAULT_STATUS_CODE
        self._http_method = http_method or DEFAULT_HTTP_METHOD

        if self.__headers__:
            for k, v in self.__headers__:
                self._headers.setdefault(k, v)

    def __call__(self, *args, **kwargs):
        abort_code = getattr(
            self, ABORT_ATTRIBUTE_NAME, None,
        )

        if abort_code:
            _abort(abort_code)

        return self.make_response()

    def __repr__(self):
        return '<ResponseProvider(url_rule={} http_method={}): {}>'.format(
            self.url_rule, self.http_method, self.endpoint,
        )

    def __eq__(self, other):
        compare = (
            self.url_rule == other.url_rule,
            self.http_method == other.http_method,
        )
        return all(compare)

    def __on_file__(self, fp):
        body = []

        for line in fp.readlines():
            try:
                line = line.decode('utf-8')
            except AttributeError:  # please python 3 :)
                pass

            if line.startswith('#'):  # pass comment line
                continue

            if self.HEADER_REGEXP.search(line):
                key, value = parse_for_header(line)
                self._headers[key] = value
            elif self.META_REGEXP.search(line):
                http_method, status_code, url_rule = parse_for_meta(line)
                self._url_rule = url_rule
                self._http_method = http_method
                self._status_code = status_code
            else:
                # Other line as body item
                body.append(line)

        self._body = u''.join(body)

    @classmethod
    def from_file(cls, file_path):
        with open(file_path, 'r') as fp:
            mock = cls(file_path, None)
            on_file(mock, fp)
        return mock

    @property
    def body(self):
        return self._body

    @property
    def headers(self):
        return self._headers

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def url_rule(self):
        return self._url_rule

    @property
    def params(self):
        if pyv.IS_PYTHON_2:
            return urlparse.parse_qs(
                urlparse.urlparse(self.url_rule).query,
            )
        return parse.parse_qs(
            parse.urlparse(self._url_rule).query,
        )

    @property
    def status_code(self):
        return int(self._status_code)

    @property
    def http_method(self):
        return self._http_method.upper()

    @property
    def mime_type(self):
        return self._mime_type or self.__mime_type__

    @property
    def content_type(self):
        return self._content_type or self.__content_type__

    @contextmanager
    def as_(self, mock):
        assert self == mock

        endpoint = self._endpoint
        self._endpoint = mock.endpoint
        try:
            yield
        finally:
            self._endpoint = endpoint

    def make_response(self):
        return Response(
            self.body,
            headers=self.headers,
            mimetype=self.mime_type,
            status=self.status_code,
            content_type=self.content_type,
        )



# -*- coding: utf-8 -*-

from warnings import warn
from optparse import OptionGroup

from . import forms
from . import polling
from .query import query
from .case import assertion
from .pageobject import Page
from .router import add_route
from .utils import re_raise_exc
from .pageobject import PageItem
from .case import require_browser
from .pageobject import PageElement
from .case import SeleniumAssertion
from .case import make_with_browsers
from .case import SeleniumCase as Case
from ...utils.common import waiting_for
from .suite import SeleniumSuite as Suite
from .browser import change_config as change_browser_config


CONFIG_KEY = 'SELENIUM_EX'


def __add_options__(parser):
    group = OptionGroup(parser, 'Selenium extension options')

    group.add_option(
        '--selenium-browser',
        dest='SELENIUM_BROWSERS',
        action='append',
        default=[],
        help='Browser name for run cases.',
    )
    group.add_option(
        '--selenium-project-url',
        dest='SELENIUM_PROJECT_URL',
        default=None,
        help='Base URL of your project.',
    )
    group.add_option(
        '--selenium-remote',
        dest='SELENIUM_REMOTE',
        action='store_true',
        default=False,
        help='Use remote server only.'
    )
    group.add_option(
        '--selenium-no-remote',
        dest='SELENIUM_NO_REMOTE',
        action='store_true',
        default=False,
        help='Use local drivers only.'
    )
    group.add_option(
        '--selenium-polling',
        dest='SELENIUM_POLLING',
        type=float,
        default=None,
        help='Polling timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-polling-delay',
        dest='SELENIUM_POLLING_DELAY',
        type=float,
        default=None,
        help='Polling delay. Float or integer value.',
    )
    group.add_option(
        '--selenium-wait-timeout',
        dest='SELENIUM_WAIT_TIMEOUT',
        type=float,
        default=None,
        help='Implicitly wait timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-page-load-timeout',
        dest='SELENIUM_PAGE_LOAD_TIMEOUT',
        type=float,
        default=None,
        help='Load page timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-script-timeout',
        dest='SELENIUM_SCRIPT_TIMEOUT',
        type=float,
        default=None,
        help='Execute script timeout. Float or integer value.',
    )
    group.add_option(
        '--selenium-window-size',
        dest='SELENIUM_WINDOW_SIZE',
        default=None,
        help='Window size. For example: 340x480',
    )

    parser.add_option_group(group)


def __install__(program):
    from .extension import EX_NAME
    from .extension import Selenium

    config = program.config.get(CONFIG_KEY, {})

    if program.config.SELENIUM_PROJECT_URL:
        config['PROJECT_URL'] = program.config.SELENIUM_PROJECT_URL

    if program.config.SELENIUM_NO_REMOTE:
        config['USE_REMOTE'] = False

    if program.config.SELENIUM_REMOTE:
        config['USE_REMOTE'] = True

    if program.config.SELENIUM_POLLING:
        config['POLLING_TIMEOUT'] = program.config.SELENIUM_POLLING

    if program.config.SELENIUM_POLLING_DELAY:
        config['POLLING_DELAY'] = program.config.SELENIUM_POLLING_DELAY

    if program.config.SELENIUM_WAIT_TIMEOUT:
        config['WAIT_TIMEOUT'] = program.config.SELENIUM_WAIT_TIMEOUT

    if program.config.SELENIUM_PAGE_LOAD_TIMEOUT:
        config['PAGE_LOAD_TIMEOUT'] = program.config.SELENIUM_PAGE_LOAD_TIMEOUT

    if program.config.SELENIUM_SCRIPT_TIMEOUT:
        config['SCRIPT_TIMEOUT'] = program.config.SELENIUM_SCRIPT_TIMEOUT

    if program.config.SELENIUM_WINDOW_SIZE:
        w, h = program.config.SELENIUM_WINDOW_SIZE.split('x')
        config['WINDOW_SIZE'] = (int(w), int(h))

    if (config.get('POLLING_TIMEOUT') and config.get('IMPLICITLY_WAIT')) \
            and config.get('POLLING_TIMEOUT') >= config.get('IMPLICITLY_WAIT'):
        warn_message = 'POLLING_TIMEOUT >= IMPLICITLY_WAIT it will be working so slowly. ' \
                       'Please set IMPLICITLY_WAIT larger POLLING_TIMEOUT'
        warn(warn_message, RuntimeWarning)

    program.shared_extension(
        EX_NAME, Selenium, args=(config, ),
    )


__all__ = (
    'Case',
    'Page',
    'forms',
    'query',
    'Suite',
    'polling',
    'PageItem',
    'add_route',
    'assertion',
    'waiting_for',
    'PageElement',
    're_raise_exc',
    'inject_driver',
    'SeleniumAssertion',
    'make_with_browsers',
    'change_browser_config',
)



# -*- coding: utf-8 -*-

import logging
from contextlib import contextmanager

from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.ext.declarative import declarative_base

from . import registry
from .constants import DEFAULT_BIND_KEY


logger = logging.getLogger(__name__)

Session = registry.get_session()


@contextmanager
def session_scope(rollback=False):
    session = Session()
    try:
        yield session
    except BaseException as error:
        if rollback:
            session.rollback()
        raise error
    finally:
        session.close()


def dict_info(params):
    """
    :type params: dict
    """
    return u', '.join([u'%s=%s' % p for p in params.items()])


class StaticProperty(property):

    def __get__(self, obj, type=None):
        return classmethod(self.fget).__get__(obj, type)()


class ModelObjects(object):

    DEFAULT_OFFSET = 0
    DEFAULT_LIMIT = 100

    def __init__(self, model):
        self.__model = model

    def get(self, pk):
        with session_scope() as session:
            obj = session.query(self.__model).get(pk)
        return obj

    def get_by(self, **params):
        with session_scope() as session:
            obj = session.query(self.__model).filter_by(**params).first()
        return obj

    def getlist(self, offset=DEFAULT_OFFSET, limit=DEFAULT_LIMIT):
        with session_scope() as session:
            result = session.query(self.__model).offset(offset).limit(limit).all()
        return result

    def getlist_by(self, offset=DEFAULT_OFFSET, limit=DEFAULT_LIMIT, **params):
        with session_scope() as session:
            result = session.query(self.__model).filter_by(**params).offset(offset).limit(limit).all()
        return result

    def update_by(self, by, **params):
        with session_scope(rollback=True) as session:
            objects = session.query(self.__model).filter_by(**by).all()

        for obj in objects:
            for k, v in params.items():
                setattr(obj, k, v)

            with session_scope() as session:
                session.add(obj)
                session.commit()
                session.refresh(obj)

        return objects

    def remove_by(self, **params):
        with session_scope() as session:
            objects = session.query(self.__model).filter_by(**params).all()

        for obj in objects:
            with session_scope(rollback=True) as session:
                session.delete(obj)
                session.commit()


class ModelCRUD(object):

    query = Session.query_property()

    def __init__(self, **params):
        logger.debug('create new object of model {}'.format(self.__class__.__name__))

        for k, v in params:
            setattr(self, k, v)

    @StaticProperty
    def objects(cls):
        return ModelObjects(cls)

    @classmethod
    @contextmanager
    def bind_key(cls, bind_key):
        to_restore = cls.__mapper__.mapped_table.info['bind_key']
        cls.__mapper__.mapped_table.info['bind_key'] = bind_key
        try:
            yield
        finally:
            cls.__mapper__.mapped_table.info['bind_key'] = to_restore

    @classmethod
    def create(cls, **params):
        instance = cls(**params)

        with session_scope(rollback=True) as session:
            session.add(instance)
            session.commit()
            session.refresh(instance)

        return instance

    def to_dict(self):
        return dict(
            (k, self.__dict__[k])
            for k in self.__dict__
            if not k.startswith('_')
        )

    def update(self, **params):
        with session_scope() as session:

            for k, v in params.items():
                setattr(self, k, v)

            session.add(self)
            session.commit()
            session.refresh(self)

    def remove(self):
        with session_scope() as session:
            session.delete(self)
            session.commit()

    def __repr__(self):
        if hasattr(self, 'id'):
            return '<{} id={}>'.format(self.__class__.__name__, self.id or 'NULL')

        return '<{}>'.format(self.__class__.__name__)


class BoundDeclarativeMeta(DeclarativeMeta):

    def __init__(self, name, bases, d):
        DeclarativeMeta.__init__(self, name, bases, d)

        try:
            bind_key = d.pop('__bind_key__', DEFAULT_BIND_KEY)
            self.__table__.info['bind_key'] = bind_key
        except AttributeError:
            pass


BaseModel = declarative_base(cls=ModelCRUD, metaclass=BoundDeclarativeMeta)



# -*- coding: utf-8 -*-

import os
import json
import logging
from functools import wraps

try:
    from httplib import HTTPException
except ImportError:  # please python 3
    from http.client import HTTPException

from selenium.common.exceptions import StaleElementReferenceException

from ... import case
from ... import steps
from ... import reason
from ... import runnable
from ...utils import pyv
from .extension import EX_NAME
from .utils import random_file_name
from ...utils.common import waiting_for


logger = logging.getLogger(__name__)


def make_with_browsers(*browsers):
    def wrapper(cls):
        setattr(cls, '__browsers__', browsers)
        return cls
    return wrapper


def require_browser(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        selenium = self.ext(EX_NAME)
        return f(
            self, selenium.browser, *args, **kwargs
        )
    return wrapper


def make_setup(case, browser_name=None):
    selenium = case.ext(EX_NAME)
    selenium.start(browser_name)

    if case.__page_class__:
        case.checkout_page(case.__page_class__)


def save_logs(case):
    selenium = case.ext(EX_NAME)
    log_file_name = lambda lt: '{}-{}({}:{}).log'.format(
        selenium.browser_name,
        lt,
        runnable.class_name(case),
        runnable.method_name(case),
    )

    def log_to_string(log):
        def create_string_from_item(item, tab=0, break_line=False):
            if isinstance(item, pyv.basestring):
                try:
                    item = json.loads(item)
                except BaseException:
                    pass

            if isinstance(item, dict):
                return ('\n' if break_line else '') + u'\n'.join(
                    u'{}{}: {}'.format(
                        ('  ' * tab), k, create_string_from_item(v, tab=(tab + 1), break_line=True)
                    ) for k, v in item.items()
                )
            elif isinstance(item, (list, tuple)):
                return ('\n' if break_line else '') + u'\n'.join(
                    u'{}{}'.format(
                        ('  ' * tab), create_string_from_item(i, tab=(tab + 1), break_line=True),
                    ) for i in item
                )
            else:
                item = pyv.unicode_string(item)

            return item.strip()

        string = create_string_from_item(log).strip()

        if pyv.IS_PYTHON_2:
            return string.encode('utf-8')
        return string

    with selenium.browser.disable_polling():
        for log_type in selenium.browser.log_types:
            log = selenium.browser.get_log(log_type)
            if not log:
                continue

            log_file_path = os.path.join(
                selenium.config.get('LOGS_PATH'),
                log_file_name(log_type),
            )

            try:
                with open(log_file_path, 'w') as f:
                    f.write(log_to_string(log))
            except BaseException as error:
                logger.error(error, exc_info=True)


class SeleniumAssertion(case.AssertionBase):

    @staticmethod
    def any_text_exist(proxy, texts, msg=None, timeout=None):
        def check_exist():
            page_text = proxy.text
            for text in texts:
                if text in page_text:
                    return True
            return False

        waiting_for(
            check_exist,
            exc_cls=AssertionError,
            message=msg or u'Not of those texts was not found on page by URL "{}". Texts: [{}]'.format(
                proxy.browser.current_url,
                u', '.join(texts),
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def text_exist(proxy, text, msg=None, timeout=None):
        def check_text():
            try:
                return text in proxy.text
            except (HTTPException, StaleElementReferenceException):
                return False

        waiting_for(
            check_text,
            exc_cls=AssertionError,
            message=msg or u'Text "{}" not found on page by URL "{}"'.format(
                text, proxy.browser.current_url,
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    def texts_exist(self, proxy, texts, timeout=None, msg=None):
        for text in texts:
            self.text_exist(proxy, text, timeout=timeout, msg=msg)

    @staticmethod
    def web_element_exist(proxy, query, msg=None, timeout=None):
        waiting_for(
            lambda: query(proxy).exist,
            exc_cls=AssertionError,
            message=msg or u'Web element was not found on page by URL "{}"'.format(
                proxy.browser.current_url,
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def web_element_not_exist(proxy, query, msg=None, timeout=None):
        waiting_for(
            lambda: not query(proxy).exist,
            exc_cls=AssertionError,
            message=msg or u'Web element was found on page by URL "{}"'.format(
                proxy.browser.current_url,
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def content_exist(proxy, query, msg=None, timeout=None):
        waiting_for(
            lambda: len(query(proxy).first().text) > 0,
            exc_cls=AssertionError,
            message=msg or 'Content is not exist inside element on page by URL "{}"'.format(
                proxy.browser.current_url,
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def content_not_exist(proxy, query, msg=None, timeout=None):
        waiting_for(
            lambda: len(query(proxy).first().text) == 0,
            exc_cls=AssertionError,
            message=msg or 'Content is exist inside element on page by URL "{}"'.format(
                proxy.browser.current_url,
            ),
            delay=proxy.config.POLLING_DELAY,
            timeout=timeout or proxy.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def current_url_equal(browser, url, msg=None, timeout=None):
        waiting_for(
            lambda: browser.current_url == url,
            exc_cls=AssertionError,
            message=msg or '{} != {}'.format(
                browser.current_url, url,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def current_url_not_equal(browser, url, msg=None, timeout=None):
        waiting_for(
            lambda: browser.current_url != url,
            exc_cls=AssertionError,
            message=msg or '{} == {}'.format(
                browser.current_url, url,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def current_path_equal(browser, path, msg=None, timeout=None):
        waiting_for(
            lambda: browser.current_path == path,
            exc_cls=AssertionError,
            message=msg or '{} != {}'.format(
                browser.current_path, path,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def current_path_not_equal(browser, path, msg=None, timeout=None):
        waiting_for(
            lambda: browser.current_path != path,
            exc_cls=AssertionError,
            message=msg or '{} == {}'.format(
                browser.current_path, path,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT,
        )

    @staticmethod
    def any_text_for_field_exist(field, texts, msg=None, timeout=None):
        def check_exist():
            page_text = field.group.area.text
            for txt in texts:
                if txt in page_text:
                    return True
            return False

        waiting_for(
            check_exist,
            exc_cls=AssertionError,
            message=msg or u'Text for field "{}" not found on page by URL "{}". Text list: [{}]'.format(
                field.name,
                field.group.area.browser.current_url,
                u', '.join(texts),
            ),
            delay=field.config.POLLING_DELAY,
            timeout=timeout or field.config.POLLING_TIMEOUT,
        )

    def texts_for_field_exist(self, field, texts, msg=None, timeout=None):
        for text in texts:
            self.text_for_field_exist(field, text, timeout=timeout, msg=msg)

    def text_for_field_exist(self, field, text, msg=None, timeout=None):
        error_message = msg or u'Text for field "{}" not found on page by URL "{}". Text: "{}"'.format(
            field.name,
            field.group.area.browser.current_url,
            text,
        )

        self.text_exist(
            field.group.area,
            text,
            timeout=timeout,
            msg=msg or error_message,
        )


assertion = SeleniumAssertion()


class SeleniumCaseLayer(case.CaseLayer):

    def on_require(self, require):
        if EX_NAME not in require:
            require.append(EX_NAME)

    def on_teardown(self, case):
        selenium = case.ext(EX_NAME)
        logs_path = selenium.config.get('LOGS_PATH')

        if logs_path:
            save_logs(case)

        selenium.stop()


class SeleniumCase(case.Case):

    __browsers__ = None
    __page_class__ = None
    __require_browser__ = True
    __layers__ = (SeleniumCaseLayer(), )
    __assertion_class__ = SeleniumAssertion

    def __init__(self, *args, **kwargs):
        if self.__require_browser__:
            kwargs.update(use_flows=False)
    
            super(SeleniumCase, self).__init__(*args, **kwargs)

            if steps.is_step_by_step_case(self):
                step_methods = []

                for step_method in steps.get_step_methods(self):
                    step_methods.append(require_browser(step_method))

                setattr(self.__class__, steps.STEPS_STORAGE_ATTRIBUTE_NAME, step_methods)
            else:
                method = getattr(self.__class__, runnable.method_name(self))
                setattr(self.__class__, runnable.method_name(self), require_browser(method))

                case.apply_flows(self)
        else:
            super(SeleniumCase, self).__init__(*args, **kwargs)

            self.page = None

    def __reason__(self):
        selenium = self.ext(EX_NAME)
        reasons = [
            super(SeleniumCase, self).__reason__(),
        ]

        try:
            screen_url = selenium.config.get('SCREEN_URL', None)
            screen_path = selenium.config.get('SCREEN_PATH', None)

            if screen_path:
                try:
                    file_name = random_file_name('.png')
                    screen_path = os.path.join(screen_path, file_name)

                    with selenium.browser.disable_polling():
                        if selenium.browser.get_screenshot_as_file(screen_path):
                            if screen_url:
                                selenium.browser.reason_storage['screen url'] = u'{}{}'.format(
                                    screen_url, file_name,
                                )
                            else:
                                selenium.browser.reason_storage['screen path'] = screen_path
                except BaseException as error:
                    logger.warn(error, exc_info=True)

            reasons.append(
                reason.item(
                    'Selenium',
                    'info from selenium extension',
                    *(u'{}: {}'.format(k, v) for k, v in selenium.browser.reason_storage.items())
                ),
            )
        except BaseException as error:
            logger.warn(error, exc_info=True)

        return reason.join(reasons)

    def __repeat__(self):
        if self.config.SELENIUM_BROWSERS and self.__repeatable__:
            for browser_name in self.config.SELENIUM_BROWSERS:
                if self.__browsers__ and browser_name not in self.__browsers__:
                    continue

                make_setup(self, browser_name)
                yield
        elif self.__browsers__ and self.__repeatable__:
            for browser_name in self.__browsers__:
                make_setup(self, browser_name)
                yield
        else:
            make_setup(self)
            yield

    def checkout_page(self, page_cls):
        self.page = page_cls(self.ext(EX_NAME).browser)



# -*- coding: utf-8 -*-

import logging
import traceback
from types import FunctionType

from . import case
from . import reason
from . import loader
from . import runnable
from .utils import pyv
from . import extensions
from .utils.common import measure_time
from .utils.common import call_to_chain
from .groups.default import DefaultCaseGroup
from .exceptions import ExtensionNotRequired
from .exceptions import ALLOW_RAISED_EXCEPTIONS


logger = logging.getLogger(__name__)


DEFAULT_LAYERS = []
MATCH_SUITE_TO_LAYER = {}


def with_match_layers(context, suite):
    for layer in context.layers:
        yield layer

    for cls, layer in MATCH_SUITE_TO_LAYER.items():
        if isinstance(suite, cls) and layer.enabled:
            yield layer


class MountData(object):

    def __init__(self, config=None):
        self.__config = config

    @property
    def config(self):
        return self.__config


class BuildRule(object):

    def __init__(self, suite_name, case_name=None, test_name=None):
        self.__suite_name = suite_name
        self.__case_name = case_name
        self.__test_name = test_name

    def __str__(self):
        if self.__suite_name and self.__case_name and self.__test_name:
            return '{}:{}.{}'.format(
                self.__suite_name, self.__case_name, self.__case_name,
            )

        if self.__suite_name and self.__case_name:
            return '{}:{}'.format(
                self.__suite_name, self.__case_name,
            )

        return self.__suite_name

    def __repr__(self):
        return '<{}(suite_name={}, case_name={}, test_name={})>'.format(
            self.__class__.__name__,
            self.__suite_name,
            self.__case_name,
            self.__test_name,
        )

    @property
    def suite_name(self):
        return self.__suite_name

    @property
    def case_name(self):
        return self.__case_name

    @property
    def test_name(self):
        return self.__test_name

    def is_of(self, suite):
        return self.__suite_name == suite.name


class SuiteLayer(runnable.LayerOfRunnableObject):

    def on_init(self, suite):
        """
        :type suite: Suite
        """
        pass

    def on_require(self, require):
        """
        :type require: list
        """
        pass

    def on_build_rule(self, build_rules):
        """
        :type build_rules: list
        """
        pass

    def on_setup(self, suite):
        """
        :type suite: Suite
        """
        pass

    def on_teardown(self, suite):
        """
        :type suite: Suite
        """
        pass

    def on_mount(self, suite, program):
        """
        :type suite: Suite
        :type program: seismograph.program.Program
        """
        pass

    def on_run(self, suite):
        """
        :type suite: Suite
        """
        pass

    def on_error(self, error, suite, result):
        """
        :type error: BaseException
        :type suite: Suite
        :type result: seismograph.result.Result
        """
        pass


class SuiteContext(runnable.ContextOfRunnableObject):

    def __init__(self, setup, teardown):
        self.__layers = []
        self.__require = []

        self.__extensions = {}
        self.__build_rules = []

        self.__setup_callbacks = [setup]
        self.__teardown_callbacks = [teardown]

    @property
    def require(self):
        return self.__require

    @property
    def extensions(self):
        return self.__extensions

    @property
    def build_rules(self):
        return self.__build_rules

    @property
    def setup_callbacks(self):
        return self.__setup_callbacks

    @property
    def teardown_callbacks(self):
        return self.__teardown_callbacks

    @property
    def layers(self):
        for layer in DEFAULT_LAYERS:
            if layer.enabled:
                yield layer

        for layer in self.__layers:
            if layer.enabled:
                yield layer

    def add_layers(self, layers):
        self.__layers.extend(layers)

    def install_extensions(self):
        for ext_name in self.__require:
            if ext_name not in self.__extensions:
                self.__extensions[ext_name] = extensions.get(ext_name)

    def start_context(self, suite):
        logger.debug(
            'Start context of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, suite), 'on_setup', suite,
            )
            call_to_chain(self.__setup_callbacks, None)
        except BaseException:
            runnable.stopped_on(suite, 'start_context')
            raise

    def stop_context(self, suite):
        logger.debug(
            'Stop context of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, suite), 'on_teardown', suite,
            )
            call_to_chain(self.__teardown_callbacks, None)
        except BaseException:
            runnable.stopped_on(suite, 'stop_context')
            raise

    def on_init(self, suite):
        logger.debug(
            'Call to chain callbacks "on_init" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        call_to_chain(
            with_match_layers(self, suite), 'on_init', suite,
        )

    def on_require(self, suite):
        logger.debug(
            'Call to chain callbacks "on_require" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        call_to_chain(
            with_match_layers(self, suite), 'on_require', self.__require,
        )

    def on_build_rule(self, suite):
        logger.debug(
            'Call to chain callbacks "on_build_rule" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        call_to_chain(
            with_match_layers(self, suite), 'on_build_rule', self.__build_rules,
        )

    def on_mount(self, suite, program):
        logger.debug(
            'Call to chain callbacks "on_mount" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        call_to_chain(
            with_match_layers(self, suite), 'on_mount', suite, program,
        )

    def on_run(self, suite):
        logger.debug(
            'Call to chain callbacks "on_run" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, suite), 'on_run', suite,
            )
        except BaseException:
            runnable.stopped_on(suite, 'on_run')
            raise

    def on_error(self, error, suite, result):
        logger.debug(
            'Call to chain callbacks "on_error" of suite "{}"'.format(
                runnable.class_name(suite),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, suite), 'on_error', error, suite, result,
            )
        except BaseException:
            runnable.stopped_on(suite, 'on_error')
            raise


class Suite(runnable.RunnableObject, runnable.MountObjectMixin, runnable.BuildObjectMixin):

    __layers__ = None
    __require__ = None
    __create_reason__ = True
    __case_class__ = case.Case
    __case_group_class__ = None
    __case_box_class__ = case.CaseBox

    #
    # Base components of runnable object
    #

    def __is_run__(self):
        return self.__is_run

    def __is_build__(self):
        return self.__is_build

    def __is_mount__(self):
        return isinstance(self.__mount_data__, MountData)

    def __class_name__(self):
        return self.__name

    def __reason__(self):
        if self.reason_storage:
            return reason.item(
                'Suite',
                'info from suite',
                *(u'{}: {}'.format(k, v) for k, v in self.reason_storage.items())
            )
        return ''

    @runnable.build_method
    def __run__(self, result):
        self.__is_run = True
        timer = measure_time()

        if result.current_state.should_stop or not self.__case_instances:
            return

        group = self._make_group()

        with result.proxy(self, timer=timer) as result_proxy:
            try:
                self.__context.on_run(self)

                with self.__context(self):
                    group(result_proxy)
            except ALLOW_RAISED_EXCEPTIONS:
                raise
            except BaseException as error:
                self.__context.on_error(error, self, result_proxy)
                result_proxy.add_error(
                    self, traceback.format_exc(), timer(), error,
                )

    #
    # Behavior on magic methods
    #

    def __iter__(self):
        return iter(self.__case_instances)

    def __nonzero__(self):
        return bool(self.__case_instances)

    def __bool__(self):  # please python 3
        return self.__nonzero__()

    #
    # Self code is starting here
    #

    def __init__(self, name, require=None, layers=None):
        super(Suite, self).__init__()

        self.__name = name

        self.__is_run = False
        self.__is_build = False

        self.__case_classes = []
        self.__case_instances = []

        self.__mount_data__ = None

        self.__context = SuiteContext(self.setup, self.teardown)

        if self.__layers__:
            self.__context.add_layers(self.__layers__)

        if layers:
            self.__context.add_layers(layers)

        if require:
            self.__context.require.extend(require)

        if self.__require__:
            self.__context.require.extend(self.__require__)

        self.__context.on_init(self)
        self.__context.on_require(self)
        self.__context.on_build_rule(self)

    def __build__(self, case_name=None, test_name=None):
        logger.debug(
            'Build suite "{}". case=name={} test_name={}'.format(
                runnable.class_name(self), case_name, test_name,
            ),
        )

        if case_name:
            cls = loader.load_case_from_suite(
                case_name, self,
            )

            self.__case_instances.extend(
                loader.load_tests_from_case(
                    cls,
                    config=self.config,
                    method_name=test_name,
                    box_class=self.__case_box_class__,
                ),
            )
        else:
            for cls in self.__case_classes:
                self.__case_instances.extend(
                    loader.load_tests_from_case(
                        cls,
                        config=self.config,
                        box_class=self.__case_box_class__,
                    ),
                )

    @property
    def name(self):
        return self.__name

    @property
    @runnable.mount_method
    def config(self):
        return self.__mount_data__.config

    @property
    def cases(self):
        return self.__case_classes

    @property
    def context(self):
        return self.__context

    def _make_group(self):
        if self.__case_group_class__:
            logger.debug(
                'Use "__case_group_class__" to making case group',
            )

            return self.__case_group_class__(
                self.__case_instances, self.config,
            )

        if self.config.GEVENT:
            logger.debug(
                'Use "GeventCaseGroup" to making case group',
            )

            pyv.check_gevent_supported()

            from .groups.gevent import GeventCaseGroup

            return GeventCaseGroup(
                self.__case_instances, self.config,
            )

        if self.config.THREADING or self.config.MULTIPROCESSING:
            logger.debug(
                'Use "ThreadingCaseGroup" to making case group',
            )

            from .groups.threading import ThreadingCaseGroup

            return ThreadingCaseGroup(
                self.__case_instances, self.config,
            )

        logger.debug(
            'Use "DefaultCaseGroup" to making case group',
        )

        return DefaultCaseGroup(
            self.__case_instances, self.config,
        )

    def setup(self, *args, **kwargs):
        pass

    def teardown(self, *args, **kwargs):
        pass

    def add_setup(self, f):
        self.__context.setup_callbacks.append(f)
        return f

    def add_teardown(self, f):
        self.__context.teardown_callbacks.append(f)
        return f

    def assign_build_rule(self, rule):
        """
        :type rule: BuildRule
        """
        assert rule.is_of(self), 'Build rule "{}" is not of this suite'.format(
            str(rule),
        )

        if rule.case_name:
            self.__context.build_rules.append(rule)

    @runnable.build_method
    def ext(self, name):
        if name not in self.__context.require:
            raise ExtensionNotRequired(name)

        return self.__context.extensions.get(name)

    def mount_to(self, program):
        logger.debug(
            'Mount suite to program "{}"'.format(
                runnable.class_name(program),
            ),
        )

        if runnable.is_mount(self):
            raise RuntimeError(
                'Suite "{}" already mount'.format(self.__name),
            )

        program.suites.append(self)

        self.__mount_data__ = MountData(
            config=program.config,
        )

        self.__context.on_mount(self, program)

    def get_map(self):
        mp = {}

        for case_class in self.__case_classes:
            mp[case_class.__name__] = {
                'cls': case_class,
                'tests': dict(
                    (atr, getattr(case_class, atr))
                    for atr in dir(case_class)
                    if atr.startswith(loader.TEST_NAME_PREFIX)
                    or
                    atr == loader.DEFAULT_TEST_NAME,
                ),
            }

        return mp

    def register(self, cls=None, **kwargs):
        if not cls and not kwargs:
            raise TypeError('cls param or **kwargs is required')
        elif cls and kwargs:
            raise TypeError('**kwargs can not be used with cls param')

        def wrapped(
                _class,
                skip=None,
                flows=None,
                layers=None,
                static=False,
                require=None,
                case_class=None,
                always_success=False,
                assertion_class=None):
            logger.debug(
                'Register case "{}.{}" on suite "{}"'.format(
                    _class.__module__, _class.__name__, self.name,
                ),
            )

            if type(_class) == FunctionType:
                _class = case.make_case_class_from_function(
                    _class,
                    static=static,
                    base_class=case_class or self.__case_class__,
                )

            if skip:
                case.skip(skip)(_class)

            if flows:
                setattr(_class, '__flows__', flows)

            if always_success:
                setattr(_class, '__always_success__', True)

            if layers:
                if _class.__layers__:
                    _class.__layers__ = tuple(_class.__layers__) + tuple(layers)
                else:
                    _class.__layers__ = tuple(layers)

            if assertion_class:
                setattr(_class, '__assertion_class__', assertion_class)

            self.__case_classes.append(
                _class.mount_to(
                    self,
                    require=require,
                ),
            )

            return _class

        if cls:
            return wrapped(cls)

        def wrapper(_class):
            return wrapped(_class, **kwargs)

        return wrapper

    @runnable.mount_method
    def build(self, case_name=None, test_name=None, shuffle=None):
        if self.__is_build:
            raise RuntimeError(
                'Suite "{}" is already built'.format(
                    self.__class__.__name__,
                ),
            )

        logger.debug(
            'Install extensions on context of suite "{}"'.format(self.name),
        )

        self.__context.install_extensions()

        if self.__context.build_rules and not case_name:
            for rule in self.__context.build_rules:
                self.__build__(
                    case_name=rule.case_name,
                    test_name=rule.test_name,
                )
        else:
            self.__build__(
                case_name=case_name,
                test_name=test_name,
            )

        if shuffle:
            shuffle(self.__case_instances)

        self.__is_build = True



# -*- coding: utf-8 -*-

import time
from functools import wraps
from types import MethodType
from collections import OrderedDict
from contextlib import contextmanager

from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.touch_actions import TouchActions
from selenium.webdriver.common.action_chains import ActionChains

from . import polling
from ...utils import pyv
from .router import Router
from .query import make_result
from ...utils.common import waiting_for
from .utils import is_ready_state_complete
from .utils import change_name_from_python_to_html


METHOD_ALIASES = {
    'go_to': 'get',
    'set': 'send_keys',
}


def factory_method(f, browser, config, allow_polling=True, reason_storage=None):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)

        if isinstance(result, WebElement):
            return WebElementProxy(
                result,
                config=config,
                browser=browser,
                allow_polling=allow_polling,
                reason_storage=reason_storage,
            )

        if isinstance(result, (list, tuple)):
            we_list = WebElementListProxy(
                browser,
                config,
                reason_storage=reason_storage,
            )

            for obj in result:
                if isinstance(obj, WebElement):
                    we_list.append(
                        WebElementProxy(
                            obj,
                            browser=browser,
                            config=config,
                            allow_polling=allow_polling,
                            reason_storage=reason_storage,
                        ),
                    )
                else:
                    return result

            return we_list
        return result
    return wrapper


class WebElementListProxy(list):

    def __init__(self, browser, config, reason_storage=None):
        super(WebElementListProxy, self).__init__()

        self.__browser = browser
        self.__config = config
        self.__reason_storage = reason_storage or OrderedDict()

    @property
    def browser(self):
        return self.__browser

    @property
    def config(self):
        return self.__config

    @property
    def reason_storage(self):
        return self.__reason_storage

    def get_by(self, **kwargs):
        try:
            return next(self.filter(**kwargs))
        except StopIteration:
            pass

    def filter(self, **kwargs):
        def check_equal(we, attr, value):
            attr_value = getattr(we, attr, None)
            if callable(attr_value) or attr_value is None:
                attr_value = getattr(we.attr, attr)
            return attr_value == value

        for we in self:
            if all(check_equal(we, k, v) for k, v in kwargs.items()):
                yield we


class BaseProxy(object):

    def __init__(self, wrapped, config=None, browser=None, reason_storage=None, allow_polling=True):
        self.__browser = browser
        self.__config = config
        self.__wrapped = wrapped
        self._allow_polling = allow_polling
        self.__reason_storage = reason_storage or OrderedDict()

        assert self.__class__ != BaseProxy, 'This is base class only. Can not be instanced.'

    def __len__(self):
        if isinstance(self._wrapped, WebElement):
            area = self
        else:
            area = self.body().first()

        try:
            return len(area.attr.innerHTML)
        except (TypeError, WebDriverException):
            return 0

    def __nonzero__(self):
        return True

    def __bool__(self):  # please python 3
        return self.__nonzero__()

    def __dir__(self):
        return list((dir(self._wrapped) + dir(self.__class__)))

    def __repr__(self):
        return repr(self._wrapped)

    def __getattr__(self, item):
        attr = getattr(
            self._wrapped, METHOD_ALIASES.get(item, item), None,
        )

        if attr is None:
            return make_result(self, item)

        if callable(attr) and type(attr) == MethodType:
            if self.allow_polling:
                return polling.do(
                    factory_method(
                        attr,
                        self.browser,
                        self.config,
                        allow_polling=self.allow_polling,
                        reason_storage=self.reason_storage,
                    ),
                    delay=self.config.POLLING_DELAY,
                    timeout=self.config.POLLING_TIMEOUT,
                )

            return factory_method(
                attr,
                self.browser,
                self.config,
                allow_polling=self.allow_polling,
                reason_storage=self.reason_storage,
            )

        return attr

    @property
    def _wrapped(self):
        return self.__wrapped

    @property
    def config(self):
        return self.__config

    @property
    def browser(self):
        return self.__browser

    @property
    def allow_polling(self):
        return (
            bool(self.config.POLLING_TIMEOUT) and self._allow_polling
        )

    @property
    def reason_storage(self):
        return self.__reason_storage

    @property
    def text(self):
        def get_text():
            if isinstance(self._wrapped, WebElement):
                return self._wrapped.text
            return self._wrapped.find_element_by_tag_name('body').text

        if self.allow_polling:
            func = polling.do(
                get_text,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return get_text()

    @property
    def is_web_element(self):
        raise NotImplementedError(
            'Property "is_web_element" does not implemented in "{}"'.format(
                self.__class__.__name__,
            ),
        )

    @contextmanager
    def do_polling(self,
                   callback,
                   exceptions=None,
                   timeout=None,
                   delay=None):
        return polling.do(
            callback,
            exceptions=exceptions or polling.POLLING_EXCEPTIONS,
            delay=delay or self.config.POLLING_DELAY,
            timeout=timeout or self.config.POLLING_TIMEOUT,
        )

    def waiting_for(self,
                    callback,
                    timeout=None,
                    exc_cls=None,
                    message=None,
                    delay=None,
                    args=None,
                    kwargs=None):
        return waiting_for(
            callback,
            args=args,
            kwargs=kwargs,
            delay=delay or self.config.POLLING_DELAY,
            timeout=timeout or self.config.POLLING_TIMEOUT,
            exc_cls=exc_cls or polling.PollingTimeoutExceeded,
            message=message or 'Wait timeout "{}" has been exceeded'.format(timeout),
        )

    @contextmanager
    def disable_polling(self):
        try:
            self._allow_polling = False
            yield
        finally:
            self._allow_polling = True

    def wait_ready(self, tries=15, delay=0.01):
        waiting_for(
            is_ready_state_complete,
            args=(self.browser, ),
            delay=self.config.POLLING_DELAY,
            timeout=self.config.POLLING_TIMEOUT,
            message='Timeout waiting for load page',
        )

        def is_update():
            length = [
                (len(self), time.sleep(delay), len(self))
                for _ in pyv.xrange(tries)
            ]
            statuses = [d[0] < d[2] for d in length]
            return True in statuses or length[0][0] != length[-1:][0][2]

        while is_update():
            time.sleep(delay)


class WebElementProxy(BaseProxy):

    def __init__(self, *args, **kwargs):
        super(WebElementProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebElement), 'This is proxy to WebElement only'

    @property
    def css(self):
        return WebElementCssToObject(self)

    @property
    def attr(self):
        return WebElementToObject(self, allow_raise=False)

    @property
    def is_web_element(self):
        return True

    def double_click(self):
        with self.browser.action_chains as action:
            action.double_click(self)

    def context_click(self):
        with self.browser.action_chains as action:
            action.context_click(self)


class WebDriverProxy(BaseProxy):

    keys = Keys

    def __init__(self, *args, **kwargs):
        super(WebDriverProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebDriver), 'This is proxy to WebDriver only'

        self.__router = Router(self)
        self.__alert = AlertProxy(self)
        self.__action_chains = ActionProxy(self, ActionChains)
        self.__touch_actions = ActionProxy(self, TouchActions)

    @property
    def browser(self):
        return self

    @property
    def router(self):
        return self.__router

    @property
    def alert(self):
        return self.__alert

    @property
    def action_chains(self):
        return self.__action_chains

    @property
    def touch_actions(self):
        return self.__touch_actions

    @property
    def current_url(self):
        if self.allow_polling:
            func = polling.do(
                lambda: self._wrapped.current_url,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return self._wrapped.current_url

    @property
    def current_path(self):
        if not self.config.PROJECT_URL:
            raise RuntimeError(
                'Can not calculate current path. PROJECT_URL is not set.',
            )
        return self.current_url.replace(self.config.PROJECT_URL, '')

    @property
    def is_web_element(self):
        return False


class AlertProxy(object):

    def __init__(self, proxy):
        self.__alert = Alert(proxy.browser)

    def __dir__(self):
        return dir(self.__alert)

    def __call__(self, proxy):
        return self.__class__(proxy.browser)

    def __getattr__(self, item):
        return getattr(self.__alert, item)

    def __repr__(self):
        return repr(self.__alert)


class ActionProxy(object):

    def __init__(self, proxy, cls):
        self.__action = cls(proxy.browser)

    def __call__(self, proxy=None):
        return self.__class__(
            proxy.browser if proxy else self.__action._driver,
            cls=self.__action.__class__,
        )

    def __dir__(self):
        return list(set(dir(self.__action) + dir(self.__class__)))

    def __getattr__(self, item):
        return getattr(self.__action, item)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        try:
            self.perform()
        finally:
            self.reset()

    def __repr__(self):
        return repr(self.__action)

    @property
    def browser(self):
        return self.__action._driver

    def reset(self):
        self.__action._actions = []


class WebElementToObject(object):

    def __init__(self, proxy, allow_raise=True):
        self.__dict__['__proxy__'] = proxy
        self.__dict__['__allow_raise__'] = allow_raise

    def __getattr__(self, item):
        atr = self.__dict__['__proxy__'].get_attribute(
            change_name_from_python_to_html(item),
        )

        if atr:
            return atr

        if self.__dict__['__allow_raise__']:
            raise AttributeError(item)

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].setAttribute(arguments[1], arguments[2]);',
            self.__dict__['__proxy__']._wrapped,
            change_name_from_python_to_html(key),
            value,
        )


class WebElementCssToObject(object):

    def __init__(self, proxy):
        self.__dict__['__proxy__'] = proxy

    def __getattr__(self, item):
        return self.__dict__['__proxy__'].value_of_css_property(
            change_name_from_python_to_html(item),
        )

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].style[arguments[1]] = arguments[2];',
            self.__dict__['__proxy__']._wrapped,
            change_name_from_python_to_html(key),
            value,
        )



# -*- coding: utf-8 -*-

import logging
from random import Random

from . import loader
from . import extensions
from .suite import BuildRule
from .exceptions import CollectError
from .utils.common import call_to_chain


logger = logging.getLogger(__name__)


def get_shuffle(config):
    if config.RANDOM:
        random = Random(config.RANDOM_SEED)
        return random.shuffle
    return None


def get_suite_name_from_command(command):
    try:
        suite_name, _ = command.split(':')
    except ValueError:
        return command

    return suite_name


def get_case_name_from_command(command):
    try:
        _, case_name = command.split(':')
        try:
            case_name, test_name = case_name.split('.')
        except ValueError:
            pass
        return case_name
    except ValueError:
        return None


def get_test_name_from_command(command):
    try:
        _, case_name = command.split(':')
        try:
            _, test_name = case_name.split('.')
            return test_name
        except ValueError:
            return None
    except ValueError:
        return None


def try_apply_rules(suite, rules):
    for rule in rules[::-1]:
        if rule.is_of(suite):
            suite.assign_build_rule(rule)
            rules.remove(rule)


def base_generator(suites, shuffle=None):
    call_to_chain(suites, 'build', shuffle=shuffle)
    extensions.clear()

    if shuffle:
        shuffle(suites)

    for suite in suites:
        yield suite


def generator_by_commands(suites, rules, shuffle=None):
    loaded_suites = []

    for rule in rules[::-1]:
        suite = loader.load_suite_by_name(rule.suite_name, suites)

        try_apply_rules(suite, rules)
        if suite not in loaded_suites:
            loaded_suites.append(suite)

    if rules:
        raise CollectError(
            'incorrect commands to collect "{}"'.format(
                ', '.join(str(r) for r in rules),
            ),
        )

    call_to_chain(loaded_suites, 'build', shuffle=shuffle)
    extensions.clear()

    if shuffle:
        shuffle(loaded_suites)

    for suite in loaded_suites:
        yield suite


def create_generator(suites, config):
    if config.TESTS:
        logger.debug('Create suite generator by commands')

        rules = [
            BuildRule(
                suite_name=get_suite_name_from_command(c),
                case_name=get_case_name_from_command(c),
                test_name=get_test_name_from_command(c),
            )
            for c in config.TESTS
        ]
        return generator_by_commands(
            suites, rules, shuffle=get_shuffle(config),
        )

    logger.debug('Create base suite generator')

    return base_generator(
        suites, shuffle=get_shuffle(config),
    )



# -*- coding: utf-8 -*-

import re
import json

from .mock import BaseMock
from .base import BaseMockServer


CLEAN_DATA_REGEXP = re.compile(r'^\s{2}|\n|\r$')
FILE_EXTENSIONS = BaseMockServer.__file_extensions__ + ('.json', )


class JsonMock(BaseMock):

    __mime_type__ = 'application/json'
    __content_type__ = 'application/json'

    @property
    def body(self):
        return json.dumps(self._body)

    @property
    def json(self):
        return self._body

    def __on_file__(self, fp):
        super(JsonMock, self).__on_file__(fp)

        # for pre validation only
        self._body = json.loads(
            CLEAN_DATA_REGEXP.sub('', self._body),
        )


class JsonApiMockServer(BaseMockServer):

    __mock_class__ = JsonMock
    __file_extensions__ = FILE_EXTENSIONS



# -*- coding: utf-8 -*-

import logging
import traceback
from functools import wraps
from unittest import TestCase as __UnitTest__

from six import with_metaclass

from . import steps
from . import loader
from . import reason
from . import runnable
from .utils import pyv
from . import extensions
from .exceptions import Skip
from .utils.common import measure_time
from .utils.common import call_to_chain
from .exceptions import ExtensionNotRequired
from .exceptions import ALLOW_RAISED_EXCEPTIONS


logger = logging.getLogger(__name__)


DEFAULT_LAYERS = []
MATCH_CASE_TO_LAYER = {}


SKIP_ATTRIBUTE_NAME = '__skip__'
SKIP_WHY_ATTRIBUTE_NAME = '__skip_why__'


def repeat(case):
    return case.__repeat__()


def repeat_method(case):
    return case.__repeat_method__()


def prepare(case, method):
    return case.__prepare__(method)


def setup_class_proxy(case):
    if getattr(case.__class__, '__setup_class_was_called__', False):
        return
    case.setup_class()
    setattr(case.__class__, '__setup_class_was_called__', True)


def teardown_class_proxy(case):
    if getattr(case.__class__, '__teardown_class_was_called__', False):
        return
    case.teardown_class()
    setattr(case.__class__, '__teardown_class_was_called__', True)


def _skip(reason):
    def wrapper(case):
        if not pyv.is_class_type(case):
            @wraps(case)
            def wrapped(*args, **kwargs):
                raise Skip(reason)
            case = wrapped

        setattr(case, SKIP_ATTRIBUTE_NAME, True)
        setattr(case, SKIP_WHY_ATTRIBUTE_NAME, reason)

        return case
    return wrapper


def skip(reason):
    return _skip(reason)


def skip_if(condition, reason):
    if condition:
        return _skip(reason)
    return lambda obj: obj


def skip_unless(condition, reason):
    if not condition:
        return _skip(reason)
    return lambda obj: obj


def set_no_skip():
    global _skip

    def no_skip(reason):
        def wrapper(f):
            return f
        return wrapper

    _skip = no_skip


def flows(*flows):
    def wrapper(f):
        if pyv.is_class_type(f):
            setattr(f, '__flows__', flows)
            return f

        @wraps(f)
        def wrapped(self, *args, **kwargs):
            for flow in flows:
                if self.config.FLOWS_LOG:
                    with self.log.tab():
                        self.log(
                            'Flow: ', pyv.unicode_string(flow),
                        )
                f(self, flow, *args, **kwargs)
        return wrapped
    return wrapper


def apply_flows(case):
    if not steps.is_step_by_step_case(case) and case.__flows__:
        setattr(
            case.__class__,
            runnable.method_name(case),
            flows(*case.__flows__)(
                getattr(case.__class__, runnable.method_name(case)),
            ),
        )


def make_case_class_from_function(
        func,
        base_class,
        doc=None,
        static=False,
        class_name=None,
        class_name_creator=None):
    if callable(class_name_creator):
        class_name = class_name_creator(func)

    if static or base_class.__static__:
        method = lambda s, *a, **k: func(*a, **k)
    else:
        method = func

    cls = type(
        class_name or func.__name__,
        (base_class, ),
        {
            '__doc__': doc or func.__doc__,
            loader.DEFAULT_TEST_NAME: method,
        },
    )

    return cls


def with_match_layers(context, case):
    for layer in context.layers:
        yield layer

    for cls, layer in MATCH_CASE_TO_LAYER.items():
        if isinstance(case, cls) and layer.enabled:
            yield layer


class CaseBox(object):

    def __init__(self, iterable):
        self.__cases = iterable
        self.__current = None

        logger.debug('CaseBox was created')

    def __call__(self, *args, **kwargs):
        return self.__run__(*args, **kwargs)

    def __iter__(self):
        for case in self.__cases:
            yield case

    def __repr__(self):
        return repr(self.__current)

    def __str__(self):
        return str(self.__current)

    def __getattr__(self, item):
        return getattr(self.__current, item)

    def __run_current__(self, result):
        if self.__current.__repeatable__ and self.__current.config.REPEAT > 0:
            for _ in pyv.xrange(self.__current.config.REPEAT):
                self.__current(result)
        else:
            self.__current(result)

    def __run__(self, result):
        for case in self.__cases:
            self.__current = case
            try:
                setup_class_proxy(self.__current)
            except BaseException as error:
                runnable.stopped_on(self.__current, 'setup_class')
                raise error
            self.__run_current__(result)

        if self.__current:
            try:
                teardown_class_proxy(self.__current)
            except BaseException as error:
                runnable.stopped_on(self.__current, 'teardown_class')
                raise error


class MountData(object):

    def __init__(self, suite_name=None, require=None):
        self.__require = require
        self.__suite_name = suite_name

    @property
    def require(self):
        return self.__require

    @property
    def suite_name(self):
        return self.__suite_name


class AssertionBase(object):

    __unittest__ = __UnitTest__('__call__')

    def len_equal(self, a, b, msg=None):
        """
        Check equal from length of list or tuple

        len_equal([1, 2, 3], 3)
        """
        if isinstance(a, (list, tuple)):
            self.equal(len(a), b, msg=msg)
        else:
            self.equal(a, len(b), msg=msg)

    def fail(self, msg=None):
        """
        Raised AssertionError with message
        """
        self.__unittest__.fail(msg)

    def true(self, expr, msg=None):
        """
        Like assertTrue in unittest
        """
        self.__unittest__.assertTrue(expr, msg=msg)

    def false(self, expr, msg=None):
        """
        Like assertFalse in unittest
        """
        self.__unittest__.assertFalse(expr, msg=msg)

    def greater(self, a, b, msg=None):
        """
        Like assertGreater in unittest
        """
        self.__unittest__.assertGreater(a, b, msg=msg)

    def equal(self, first, second, msg=None):
        """
        Like assertEqual in unittest
        """
        self.__unittest__.assertEqual(first, second, msg=msg)

    def not_equal(self, first, second, msg=None):
        """
        Like assertNotEqual in unittest
        """
        self.__unittest__.assertNotEqual(first, second, msg=msg)

    def raises(self, exc_class, callable_obj=None, *args, **kwargs):
        """
        Like assertRaises in unittest
        """
        return self.__unittest__.assertRaises(exc_class, callable_obj, *args, **kwargs)

    def is_instance(self, obj, cls, msg=None):
        """
        Like assertIsInstance in unittest
        """
        self.__unittest__.assertIsInstance(obj, cls, msg=msg)

    def sequence_equal(self, seq1, seq2, msg=None, seq_type=None):
        """
        Like assertSequenceEqual in unittest
        """
        self.__unittest__.assertSequenceEqual(seq1, seq2, msg=msg, seq_type=seq_type)

    def almost_equal(self, first, second, places=None, msg=None, delta=None):
        """
        Like assertAlmostEqual in unittest
        """
        self.__unittest__.assertAlmostEqual(first, second, places=places, msg=msg, delta=delta)

    def not_almost_equal(self, first, second, places=None, msg=None, delta=None):
        """
        Like assertNotAlmostEqual in unittest
        """
        self.__unittest__.assertNotAlmostEqual(first, second, places=places, msg=msg, delta=delta)

    def is_in(self, member, container, msg=None):
        """
        Like assertIn in unittest
        """
        self.__unittest__.assertIn(member, container, msg=msg)

    def is_not_in(self, member, container, msg=None):
        """
        Like assertNotIn in unittest
        """
        self.__unittest__.assertNotIn(member, container, msg=msg)

    def is_none(self, obj, msg=None):
        """
        Like assertIsNone in unittest
        """
        self.__unittest__.assertIsNone(obj, msg=msg)

    def is_not_none(self, obj, msg=None):
        """
        Like assertIsNotNone in unittest
        """
        self.__unittest__.assertIsNotNone(obj, msg=msg)


class CaseLayer(runnable.LayerOfRunnableObject):

    def on_init(self, case):
        """
        :type case: Case
        """
        pass

    def on_require(self, require):
        """
        :type require: list
        """
        pass

    def on_setup(self, case):
        """
        :type case: Case
        """
        pass

    def on_teardown(self, case):
        """
        :type case: Case
        """
        pass

    def on_skip(self, case, reason, result):
        """
        :type case: Case
        :type reason: str
        :type result: seismograph.result.Result
        """
        pass

    def on_any_error(self, error, case, result):
        """
        :type case: Case
        :type error: BaseException
        :type result: seismograph.result.Result
        """
        pass

    def on_error(self, error, case, result):
        """
        :type case: Case
        :type error: BaseException
        :type result: seismograph.result.Result
        """
        pass

    def on_context_error(self, error, case, result):
        """
        :type case: Case
        :type error: BaseException
        :type result: seismograph.result.Result
        """
        pass

    def on_fail(self, fail, case, result):
        """
        :type case: Case
        :type fail: AssertionError
        :type result: seismograph.result.Result
        """
        pass

    def on_success(self, case):
        """
        :type case: Case
        """
        pass

    def on_run(self, case):
        """
        :type case: Case
        """
        pass


class CaseContext(runnable.ContextOfRunnableObject):

    def __init__(self,
                 setup,
                 teardown,
                 layers=None):
        self.__require = []
        self.__extensions = {}
        self.__layers = layers if layers else []

        self.__setup_callbacks = [setup]
        self.__teardown_callbacks = [teardown]

    @property
    def require(self):
        return self.__require

    @property
    def extensions(self):
        return self.__extensions

    @property
    def setup_callbacks(self):
        return self.__setup_callbacks

    @property
    def teardown_callbacks(self):
        return self.__teardown_callbacks

    @property
    def layers(self):
        for layer in DEFAULT_LAYERS:
            if layer.enabled:
                yield layer

        for layer in self.__layers:
            if layer.enabled:
                yield layer

    def start_context(self, case):
        logger.debug(
            'Start context of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_setup', case,
            )
            call_to_chain(self.__setup_callbacks, None)
        except BaseException:
            runnable.stopped_on(case, 'start_context')
            raise

    def stop_context(self, case):
        logger.debug(
            'Stop context of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_teardown', case,
            )
            call_to_chain(self.__teardown_callbacks, None)
        except BaseException:
            runnable.stopped_on(case, 'stop_context')
            raise

    def install_extensions(self):
        for ext_name in self.require:
            if ext_name not in self.__extensions:
                self.__extensions[ext_name] = extensions.get(ext_name)

    def on_init(self, case):
        logger.debug(
            'Call to chain callbacks "on_init" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        call_to_chain(
            with_match_layers(self, case), 'on_init', case,
        )

    def on_require(self, case):
        logger.debug(
            'Call to chain callbacks "on_require" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        call_to_chain(
            with_match_layers(self, case), 'on_require', self.__require,
        )

    def on_skip(self, case, reason, result):
        logger.debug(
            'Call to chain callbacks "on_skip" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_skip', case, reason, result,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_skip')
            raise

    def on_any_error(self, error, case, result):
        logger.debug(
            'Call to chain callbacks "on_any_error" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_any_error', error, case, result,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_any_error')
            raise

    def on_error(self, error, case, result):
        logger.debug(
            'Call to chain callbacks "on_error" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_error', error, case, result,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_error')
            raise

    def on_context_error(self, error, case, result):
        logger.debug(
            'Call to chain callbacks "on_context_error" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_context_error', error, case, result,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_context_error')
            raise

    def on_fail(self, fail, case, result):
        logger.debug(
            'Call to chain callbacks "on_fail" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_fail', fail, case, result,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_fail')
            raise

    def on_success(self, case):
        logger.debug(
            'Call to chain callbacks "on_success" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_success', case,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_success')
            raise

    def on_run(self, case):
        logger.debug(
            'Call to chain callbacks "on_run" of case "{}"'.format(
                runnable.class_name(case),
            ),
        )

        try:
            call_to_chain(
                with_match_layers(self, case), 'on_run', case,
            )
        except BaseException:
            runnable.stopped_on(case, 'on_run')
            raise


assertion = AssertionBase()


class Case(with_metaclass(steps.CaseMeta, runnable.RunnableObject, runnable.MountObjectMixin)):

    __flows__ = None
    __layers__ = None
    __static__ = False
    __require__ = None
    __repeatable__ = True
    __create_reason__ = True
    __always_success__ = False
    __assertion_class__ = None

    #
    # Base components of runnable object
    #

    def __is_run__(self):
        return self.__is_run

    def __is_mount__(self):
        mount_data = getattr(
            self, '__mount_data__', None,
        )
        return isinstance(mount_data, MountData)

    def __method_name__(self):
        return self._method_name

    def __class_name__(self):
        return '{}.{}'.format(
            self.__mount_data__.suite_name, self.__class__.__name__,
        )

    def __reason__(self):
        reasons = []

        if steps.is_step_by_step_case(self):
            history = steps.get_case_history(self) or [None]

            reasons.append(
                reason.join(
                    (
                        reason.item(
                            'History',
                            'was done earlier',
                            *history
                        ),
                        reason.item(
                            'Current step',
                            'when exception was raised',
                            steps.get_current_step(self),
                        ),
                        reason.item(
                            'Current flow',
                            'context of steps execution',
                            steps.get_current_flow(self),
                        ),
                    ),
                ),
            )

        if self.reason_storage:
            reasons.append(
                reason.item(
                    'Case',
                    'info from test case',
                    *(u'{}: {}'.format(k, v) for k, v in self.reason_storage.items())
                ),
            )

        return reason.join(*reasons)

    def __run__(self, result):
        self.__is_run = True
        timer = measure_time()

        if result.current_state.should_stop:
            return

        with result.proxy() as result_proxy:
            result_proxy.start(self)

            if self.__always_success__:
                self.__context.on_success(self)
                result_proxy.add_success(
                    self, timer(),
                )
                return

            if hasattr(self, SKIP_ATTRIBUTE_NAME):
                reason = getattr(self, SKIP_WHY_ATTRIBUTE_NAME, 'no reason')
                self.__context.on_skip(self, reason, result_proxy)
                result_proxy.add_skip(
                    self, reason, timer(),
                )
                return

            self.__log = result_proxy.console.child_console()

            try:
                self.__context.on_run(self)

                was_success = True

                for _ in iter(repeat(self)):
                    with self.__context(self):
                        try:
                            test_method = prepare(
                                self, getattr(self, runnable.method_name(self)),
                            )
                            for _ in iter(repeat_method(self)):
                                test_method()
                        except ALLOW_RAISED_EXCEPTIONS:
                            result_proxy.current_state.should_stop = True
                            raise
                        except Skip as s:
                            was_success = False
                            self.__context.on_skip(self, s.message, result_proxy)
                            result_proxy.add_skip(
                                self, s.message, timer(),
                            )
                        except AssertionError as fail:
                            was_success = False
                            self.__context.on_fail(fail, self, result_proxy)
                            result_proxy.add_fail(
                                self, traceback.format_exc(), timer(), fail,
                            )
                        except BaseException as error:
                            was_success = False
                            self.__context.on_error(error, self, result_proxy)
                            self.__context.on_any_error(error, self, result_proxy)
                            result_proxy.add_error(
                                self, traceback.format_exc(), timer(), error,
                            )

                    if not was_success:
                        break

                if was_success:
                    self.__context.on_success(self)
                    result_proxy.add_success(
                        self, timer(),
                    )
            except ALLOW_RAISED_EXCEPTIONS:
                raise
            except BaseException as error:
                self.__context.on_context_error(error, self, result_proxy)
                self.__context.on_any_error(error, self, result_proxy)
                result_proxy.add_error(
                    self, traceback.format_exc(), timer(), error,
                )

    #
    # Behavior on magic methods
    #

    def __str__(self):
        return '{} ({}:{})'.format(
            self._method_name,
            self.__mount_data__.suite_name,
            self.__class__.__name__,
        )

    def __repr__(self):
        class_path = '{}:{}'.format(
            self.__mount_data__.suite_name, self.__class__.__name__,
        )
        return '<{} method_name={} stopped_on={}>'.format(
            class_path, runnable.method_name(self), runnable.stopped_on(self),
        )

    #
    # Self code is starting here
    #

    @runnable.mount_method
    def __init__(self, method_name, config=None, use_flows=True):
        if not hasattr(self, method_name):
            raise AttributeError(
                '"{}" does not have attribute "{}"'.format(
                    self.__class__.__name__,
                    method_name,
                ),
            )

        self.__log = None
        self.__is_run = False
        self.__config = config
        self._method_name = method_name

        if use_flows:
            apply_flows(self)

        self.__context = CaseContext(
            self.setup,
            self.teardown,
            layers=self.__layers__,
        )

        if self.__mount_data__.require:
            self.__context.require.extend(
                self.__mount_data__.require,
            )

        if self.__require__:
            self.__context.require.extend(
                self.__require__,
            )

        if self.__assertion_class__:
            self.__assertion = self.__assertion_class__()
        else:
            self.__assertion = assertion

        self.__context.on_init(self)
        self.__context.on_require(self)

        logger.debug(
            'Install extensions on context of case "{}"'.format(
                runnable.class_name(self),
            ),
        )

        self.__context.install_extensions()

        super(Case, self).__init__()

    @classmethod
    def mount_to(cls, suite, require=None):
        logger.debug(
            'Mount case "{}.{}" to suite "{}"'.format(
                cls.__module__, cls.__name__, suite.name,
            ),
        )

        if getattr(cls, '__mount_data__', None) is not None:
            raise RuntimeError(
                'Case "{}" already mounted'.format(cls.__name__),
            )

        common_require = []

        if require:
            common_require.extend(require)

        if suite.context.require:
            common_require.extend(suite.context.require)

        cls.__mount_data__ = MountData(
            suite_name=suite.name,
            require=common_require,
        )

        return cls

    def __repeat__(self):
        yield

    def __repeat_method__(self):
        yield

    def __prepare__(self, method):
        return method

    @property
    def config(self):
        return self.__config

    @property
    @runnable.run_method
    def log(self):
        return self.__log

    @property
    def context(self):
        return self.__context

    @property
    def assertion(self):
        return self.__assertion

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup(self, *args, **kwargs):
        pass

    def teardown(self, *args, **kwargs):
        pass

    def ext(self, name):
        if name not in self.__context.require:
            raise ExtensionNotRequired(name)

        return self.__context.extensions.get(name)

    @runnable.run_method
    def skip_test(self, reason):
        _skip(reason)(lambda: None)()
